@@ -1,21 +1,8 @@
+// SPDX-License-Identifier: GPL-2.0-or-later
 /*
  * cryptsetup crypto name and hex conversion helper test vectors
  *
- * Copyright (C) 2022-2024 Milan Broz
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ * Copyright (C) 2022-2025 Milan Broz
  */
 
 #include <stdio.h>
@@ -121,7 +108,7 @@ static int test_parse_integrity_mode(void)
 		key_size = -1;
 		memset(integrity, 0, sizeof(integrity));
 		if (integrity_test_vectors[i].int_mode &&
-		    (crypt_parse_integrity_mode(integrity_test_vectors[i].input, integrity, &key_size) < 0 ||
+		    (crypt_parse_integrity_mode(integrity_test_vectors[i].input, integrity, &key_size, 0) < 0 ||
 			strcmp(integrity_test_vectors[i].integrity, integrity) ||
 			integrity_test_vectors[i].key_size != key_size)) {
 			printf("[FAILED (%s / %i)]\n", integrity, key_size);
