@@ -1,25 +1,12 @@
+// SPDX-License-Identifier: GPL-2.0-or-later
 /*
  * Example of LUKS2 token storing third party metadata (EXPERIMENTAL EXAMPLE)
  *
- * Copyright (C) 2016-2021 Milan Broz <gmazyland@gmail.com>
- * Copyright (C) 2021 Vojtech Trefny
+ * Copyright (C) 2016-2025 Milan Broz
+ * Copyright (C) 2021-2025 Vojtech Trefny
  *
  * Use:
  *  - generate ssh example token
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include <stdio.h>
@@ -47,6 +34,7 @@
 #define OPT_DEBUG	5
 #define OPT_DEBUG_JSON	6
 #define OPT_KEY_SLOT	7
+#define OPT_TOKENS_PATH	8
 
 void tools_cleanup(void)
 {
@@ -59,6 +47,7 @@ static int token_add(
 		const char *user,
 		const char *path,
 		const char *keypath,
+		const char *plugin_path,
 		int keyslot)
 
 {
@@ -68,23 +57,36 @@ static int token_add(
 	const char *string_token;
 	int r, token;
 
+	if (plugin_path) {
+		r = crypt_token_set_external_path(plugin_path);
+		if (r < 0)
+			return r;
+	}
+
 	r = crypt_init(&cd, device);
 	if (r)
 		return r;
 
 	r = crypt_load(cd, CRYPT_LUKS2, NULL);
-	if (r)
+	if (r) {
+		l_err(cd, _("Device %s is not a valid LUKS device."), device);
 		goto out;
+	}
 
-	r = -EINVAL;
 	jobj = json_object_new_object();
-	if (!jobj)
+	if (!jobj) {
+		r = -ENOMEM;
 		goto out;
+	}
 
 	/* type is mandatory field in all tokens and must match handler name member */
 	json_object_object_add(jobj, "type", json_object_new_string(TOKEN_NAME));
 
 	jobj_keyslots = json_object_new_array();
+	if (!jobj_keyslots) {
+		r = -ENOMEM;
+		goto out;
+	}
 
 	/* mandatory array field (may be empty and assigned later */
 	json_object_object_add(jobj, "keyslots", jobj_keyslots);
@@ -141,6 +143,8 @@ static struct argp_option options[] = {
 	{"ssh-user",	OPT_SSH_USER, 	"STRING", 0, N_("Username used for the remote server")},
 	{"ssh-path",	OPT_SSH_PATH,	"STRING", 0, N_("Path to the key file on the remote server")},
 	{"ssh-keypath",	OPT_KEY_PATH, 	"STRING", 0, N_("Path to the SSH key for connecting to the remote server")},
+	{"external-tokens-path",
+			OPT_TOKENS_PATH,"STRING", 0, N_("Path to directory containinig libcryptsetup external tokens")},
 	{"key-slot",	OPT_KEY_SLOT,	"NUM",	  0, N_("Keyslot to assign the token to. If not specified, token will "\
 						        "be assigned to the first keyslot matching provided passphrase.")},
 	{0,		0,		0,	  0, N_("Generic options:")},
@@ -157,6 +161,7 @@ struct arguments {
 	char *ssh_user;
 	char *ssh_path;
 	char *ssh_keypath;
+	char *ssh_plugin_path;
 	int keyslot;
 	int verbose;
 	int debug;
@@ -180,6 +185,9 @@ parse_opt (int key, char *arg, struct argp_state *state) {
 	case OPT_KEY_PATH:
 		arguments->ssh_keypath = arg;
 		break;
+	case OPT_TOKENS_PATH:
+		arguments->ssh_plugin_path = arg;
+		break;
 	case OPT_KEY_SLOT:
 		arguments->keyslot = atoi(arg);
 		break;
@@ -406,6 +414,7 @@ int main(int argc, char *argv[])
 				arguments.ssh_user,
 				arguments.ssh_path,
 				arguments.ssh_keypath,
+				arguments.ssh_plugin_path,
 				arguments.keyslot);
 		if (ret < 0)
 			return EXIT_FAILURE;
